/* 
 * All rights reserved.
 *
 * Common parts of TELNET and FTP shells.
 *
 * This file is part of the uIP TCP/IP stack
 *
 * $Id: telnetd.c,v 1.2 2006/06/07 09:43:54 adam Exp $
 * Modified by Jaroslav Fojtik 2010 - 2019
 */
#include <string.h>

#include "../misc/memb.h"

#include "tel_tools.h"

#include "uip.h"


//#include "logger.h"	/// !!!!! To be removed


typedef struct {
  char line[CONF_LINELEN_OUT];
} ftpd_line;
MEMB(linemem, sizeof(ftpd_line), CONF_NUMLINES);



#ifndef dealloc_line
void dealloc_line(char *line)
{
  if(LOCATED_IN_RAM(line))
     memb_free(linemem, line);
}
#endif


#ifndef alloc_line
char *alloc_line(void)
{
  return memb_alloc(linemem);
}
#endif


#ifdef CONF_NUMLINES
 #define NUM_LINES CONF_NUMLINES
int sendlineL(char *lines[], char *line, unsigned char TmpNumsent)
#else
int sendline(char *lines[], char *line, unsigned char TmpNumsent, int NUM_LINES)
#endif
{
  while(TmpNumsent<NUM_LINES) 
  {
    if(lines[TmpNumsent] == NULL) 
    {
      lines[TmpNumsent] = line;
      return 0;
    }
    TmpNumsent++;
  }

   	/* No line to store. */
  dealloc_line(line);
  return -1;
}


/** After TCP/IP returns ACK, deallocate all lines sent. */
#ifdef CONF_NUMLINES
void ackedL(char *lines[], unsigned char *numsent)
#else
void acked(char *lines[], unsigned char *numsent, int NUM_LINES)
#endif
{
unsigned int i;

/* faster code replacement to test 
  if(*numsent>NUM_LINES) *numsent=NUM_LINES;
  i = *numsent;
  while(i>0)
  {
    i--;
    dealloc_line(lines[i]);    
  }
					// numsent lines erased 
  i = NUM_LINES - *numsent;  // i lines must be shifted down
  memcpy(lines[0], lines[*numsent], i*sizeof(char*));
  memset(lines[i], 0, *numsent*sizeof(char*));
  *numsent = 0;        */

  for(i=0; i<*numsent; i++)
  {
    dealloc_line(lines[i]);
    lines[i] = 0;
  }
  *numsent = 0;
  while(i < NUM_LINES)
  {
    if(lines[i] != NULL)
    {
      if(i != *numsent)
      {
        lines[*numsent] = lines[i];
        lines[i] = NULL;
      }
      (*numsent)++;
    }
    i++;
  }
  *numsent = 0;
}


/** Store char into a local line buff
 * param[out]	 buf	Buffer to store new character.
 * param[in,out] bufpos	Current position in buffer buf.
 * param[in]	 c	Character read.
 * param[in]	 bufsize Maximal extent of the buffer. */
int store_char(char *buf, unsigned char *bufpos, unsigned char c, unsigned bufsize)
{
  if(c==ISO_nl || c==0)		/* Ignore Nl and 0 character. */
  {
    return 0;
  }

  if(c == 0x8)			/* Backspace */
  {
    if(*bufpos>0) (*bufpos)--;
    return 0;
  }

  if(c==ISO_cr || *bufpos>=bufsize-1) 
  {
    buf[*bufpos] = 0;		/* append trailing 0. */        
    return 1;
  }
  
  buf[*bufpos] = c;
  ++*bufpos;
  return 0;
}


#ifdef CONF_NUMLINES
void sendopt(char *lines[], unsigned char option, unsigned char value, unsigned char TmpNumsent)
#else
void sendopt(char *lines[], unsigned char option, unsigned char value, unsigned char TmpNumsent, int NUM_LINES)
#endif
{
char *line;
  line = alloc_line();
  if(line != NULL)
  {
    line[0] = IAC_IAC;
    line[1] = option;
    line[2] = value;
    line[3] = 0;
    sendline(lines,line,TmpNumsent,NUM_LINES);
  }
}


#ifdef CONF_NUMLINES
void close_lines(char *lines[])
#else
void close_lines(char *lines[], int NUM_LINES)
#endif
{
unsigned int i;
  
  for(i=0; i<NUM_LINES; ++i)
  {
    if(lines[i] != NULL)
    {
      dealloc_line(lines[i]);
      lines[i] = NULL;
    }
  }
}


/** Send all data from numsent position till the end. */
#ifdef CONF_NUMLINES
int senddata(char *lines[], unsigned char *numsent)
#else
int senddata(char *lines[], unsigned char *numsent, int NUM_LINES)
#endif
{
char *bufptr, *lineptr;
int buflen, linelen;
unsigned i;
  
  bufptr = uip_appdata;
  buflen = 0;
  
  i = *numsent;
  while(i<NUM_LINES) 
  {
    lineptr = lines[i];
    if(lineptr==NULL)
    {
      i++;
      continue;
    }

    linelen = strlen(lineptr);
    if(linelen>CONF_LINELEN_OUT)
	{linelen = CONF_LINELEN_OUT;}

    if(buflen + linelen >= uip_mss()) break;
    
    memcpy(bufptr, lineptr, linelen);
    //LogEvent(DEBUG_EVENT, "Send Telnet: '%s'", lineptr);

    bufptr += linelen;
    buflen += linelen;    
    i++;
    *numsent = i;	// Assign *numsent behind non NULL row only.
  }

  if(buflen > 0)
    uip_send(uip_appdata, buflen);
  return buflen;
}



/** Interpret As Command - commands. */
#ifdef CONF_NUMLINES
void IAC(char *lines[], unsigned char c, term_state *TS, unsigned char TmpNumsent)
#else
void IAC(char *lines[], unsigned char c, term_state *TS, unsigned char TmpNumsent, int NUM_LINES)
#endif
{
  if(TS==NULL) return;
  switch(TS->state)
  {
    case STATE_IAC:
       if(c == IAC_IAC) 
       {
	 TS->state = STATE_NORMAL;
         return;
       }

       switch(c) 
       {
	 case IAC_WILL: TS->state = STATE_WILL; return;
         case IAC_WONT: TS->state = STATE_WONT; return;
         case IAC_DO:   TS->state = STATE_DO; return;
	 case IAC_DONT: TS->state = STATE_DONT; return;
	 /*default:       return STATE_AFTER_IAC;*/
       }      
       TS->state = STATE_AFTER_IAC;
       return;

    case STATE_WILL:		/* Reply with a DONT */
      if(c!=TS->ActiveCommand) sendopt(lines, IAC_DONT, c, TmpNumsent);
      if(c==IAC_TERM_TYPE) TS->InteligentTerm = 1;
      TS->state = STATE_AFTER_IAC;
      TS->ActiveCommand = 0xFF;
      return;
      
    case STATE_WONT:		/* Reply with a DONT */
      if(c!=TS->ActiveCommand) sendopt(lines, IAC_DONT, c, TmpNumsent);
      TS->state = STATE_AFTER_IAC;
      TS->ActiveCommand = 0xFF;
      return;

    case STATE_DO:		/* Reply with a WONT */
      if(c!=TS->ActiveCommand) sendopt(lines, IAC_WONT, c, TmpNumsent);
      TS->state = STATE_AFTER_IAC;
      TS->ActiveCommand = 0xFF;
      return;

    case STATE_DONT:		/* Reply with a WONT */
      if(c!=TS->ActiveCommand) sendopt(lines, IAC_WONT, c, TmpNumsent);
      TS->state = STATE_AFTER_IAC;
      TS->ActiveCommand = 0xFF;
      return;

    default:			/* unknown state */
    case STATE_AFTER_IAC:
      
    case STATE_NORMAL:
      if(c == IAC_IAC)
      {
	TS->state = STATE_IAC;
        return;
      }
      TS->state = STATE_NORMAL;
      return;

    case STATE_CLOSE:
      TS->state = STATE_CLOSE;
      return;
    }
}


/** Print two strings to the shell window. 
 * This function is implemented by the shell GUI / telnet server and
 * can be called by the shell back-end to output a string in the
 * shell window. The string is automatically appended with a linebreak. 
 * @param str1		The first half of the string to be output.
 * @param str2		The second half of the string to be output. */
#ifdef CONF_NUMLINES
void shell_output(char *lines[], const char *str1, const char *str2, unsigned char TmpNumsent)
#else
void shell_output(char *lines[], const char *str1, const char *str2, unsigned char TmpNumsent, int NUM_LINES)
#endif
{
unsigned len;
char *line;

  //LogEvent(DEBUG_EVENT, "shell_output: '%s' '%s'", str1, str2);
  line = ftp_alloc_line();
  if(line != NULL) {
    len = strlen(str1);
    strncpy(line, str1, CONF_LINELEN_OUT);
    if(len < CONF_LINELEN_OUT) 
    {
      if(str2 != NULL)
        strncpy(line + len, str2, CONF_LINELEN_OUT - len);
    }
    len = strlen(line);
    if(len < CONF_LINELEN_OUT - 3)
    {
      line[len] = ISO_cr;
      line[len+1] = ISO_nl;
      line[len+2] = 0;
    }    
    sendline(lines,line,TmpNumsent,CONF_NUMLINES);
  }
}


/** Print a strings to the shell window. 
 * This function is implemented by the shell GUI / telnet server and
 * can be called by the shell back-end to output a string in the
 * shell window. The string is automatically appended with a linebreak. 
 * @param str		The string to be output.*/
#ifdef CONF_NUMLINES
int shell_output1(char *lines[], const char *str, unsigned char TmpNumsent)
#else
int shell_output1(char *lines[], const char *str, unsigned char TmpNumsent, int NUM_LINES)
#endif
{
char *line;
unsigned len;

  //LogEvent(DEBUG_EVENT, "shell_output1: '%s'", str);
  if(str==NULL) return 0;

  len = strlen(str);

  while(len > CONF_LINELEN_OUT-2)
  {   
    line = ftp_alloc_line();
    if(line == NULL) return -10;

    memcpy(line, str, CONF_LINELEN_OUT-2);
    line[CONF_LINELEN_OUT-1] = 0;

    if(sendline(lines, line, TmpNumsent, CONF_NUMLINES-1) <0)
      return -11;

    str += CONF_LINELEN_OUT-2;
    len -= CONF_LINELEN_OUT-2;
  }

  line = ftp_alloc_line();
  if(line == NULL) return -12;

  memcpy(line, str, len);
  if(len<CONF_LINELEN_OUT-3)
  {
    line[len] = ISO_cr;
    line[len+1] = ISO_nl;
    line[len+2] = 0;
  }
  return sendline(lines, line, TmpNumsent, CONF_NUMLINES);
}

