/******************************************************************************
* *
* evolving systems consulting *
* http://www.evolvsys.cz *
* *
* ALL RIGHTS RESERVED *
 *									      *
******************************************************************************/
/** @file PktPool.c Packet pool.
 *  Dependency HW: No
 *  Dependency SW: IAR or KEIL, Free RTOS
 *
 * $HeadURL: svn://213.165.82.37/UAV_AUTOPILOT/SW/src/drivers/PktPool.c $
 * $Revision: 663 $
 * $Date: 2013-01-22 11:50:10 +0100 (út, 22 I 2013) $
 * $Author: escsro026 $ */


/*------------------------------------------------------------------------------
 *				Includes
 *----------------------------------------------------------------------------*/

  // common includes
#include <string.h>

  // FreeRTOS includes
#include "FreeRTOS.h"
#include "task.h"
#include "queue.h"

  /* specific includes */
#include "typedfs.h"


#include "uipopt.h"

#ifdef emacCONF_NUM_RX_BUFFERS
  #define emacNUM_RX_BUFFERS emacCONF_NUM_RX_BUFFERS
#else
 #define emacNUM_RX_BUFFERS		   6
#endif  /* emacNUM_RX_BUFFERS */


/*------------------------------------------------------------------------------
 *				Variables
 *----------------------------------------------------------------------------*/


#define emacFRAM_SIZE_BYTES 		   2
#define emacFRAM_PADDING 		   6

	// !!!!! \todo proc je tady 4* ????. Kdyz se sem nenapise 4x, ale jen 2x tak to padne na zobak.
static UINT8 PktBuffers[emacNUM_RX_BUFFERS][UIP_BUFSIZE + emacFRAM_SIZE_BYTES + emacFRAM_PADDING]
#if defined(__ARMCC_VERSION)
 		__attribute__ ((aligned (4)))
#endif
		;
static volatile UINT8 PktAllocMap[(emacNUM_RX_BUFFERS+7)/8]; ///< Allocation map for ::CAN_X_Buffer.

/** Nibble fast grid allocation. */
extern const INT8 ALLOC_FREE[16];


xQueueHandle xEthRxInterruptQueue = NULL;
xQueueHandle xEthTxInterruptQueue = NULL;


/*------------------------------------------------------------------------------
 *				Functions
 *----------------------------------------------------------------------------*/

/** This procedure gets a new packet from packet cache.
 * Please note that the packet must be released by a function ::ReleaseMsgBox.
 * @param[in]	CallFromIRQ	0 - no IRQ; 1 - called from IRQ.
 * @return      Pointer to a packet structure; NULL when a cache is exhausted.*/
unsigned char *GetFreePkt(char CallFromIRQ)
{
INT8 i,j;

	/* Fast allocation tree scan */
  for(i=0;i<sizeof(PktAllocMap);i++)
  {
    if(PktAllocMap[i]!=0xFF)
    {
LOOK_AGAIN:
      j = ALLOC_FREE[PktAllocMap[i] & 0xF];		//try first nibble
      if(j>=32)
      {
        j = ALLOC_FREE[PktAllocMap[i] >> 4] + 4;	//try second nibble
	if(j>=32) continue;
      }

		/* The critical section ensures that only one task could allocate one mailbox. */
      if(CallFromIRQ)
      {
	PktAllocMap[i] |= (1<<j);
      }
      else
      {
        vPortEnterCritical();
        if(PktAllocMap[i] & (1<<j))
        { 						// somebody steal asynchronously our mailbox.
          vPortExitCritical();
          goto LOOK_AGAIN;				// continue in looking for free mailbox
        }
        PktAllocMap[i] |= (1<<j);			// mark index as used
        vPortExitCritical();
      }
		/* End of critical piece of code. */

      j += 8*i;						// free cache index found
      if(j>=emacNUM_RX_BUFFERS) return NULL;
      //if(!CallFromIRQ) LogEvent(DEBUG_EVENT,"Box %p allocated.", &CAN_X_Buffer[j]);
      return PktBuffers[j];
    }
  }
  return NULL;		// No Mailbox available
}


/** This function disposes packet structure.
  * @param[in]	pPacket  Pointer to a packet to be dispossed.
  * @param[in]	CallFromIRQ	0 - no IRQ; 1 - called from IRQ. */
void DisposePkt(unsigned char *pPacket, char CallFromIRQ)
{
size_t off;
  if(pPacket==NULL || 
     pPacket<PktBuffers[0]) return;	/* Packet is before Buffer. */
  off = (pPacket - PktBuffers[0])/(UIP_BUFSIZE + (4*emacFRAM_SIZE_BYTES));
  if(off>=emacNUM_RX_BUFFERS)
  {
    return;
  }

  if(!CallFromIRQ) vPortEnterCritical();
  PktAllocMap[off/8] &= ~(1<<(off%8));
  if(!CallFromIRQ) vPortExitCritical();
}


unsigned GetPktNum(unsigned char *pPacket)
{
  return (pPacket - PktBuffers[0])/(UIP_BUFSIZE + (4*emacFRAM_SIZE_BYTES));
}


/** This function initialises whole MESSAGE pool. */
void InitPktPool(void)
{
char ulTemp;

  memset((void *)PktAllocMap,0,sizeof(PktAllocMap));

  	/* Mark each Rx buffer as empty. */
  for(ulTemp=0; ulTemp<emacNUM_RX_BUFFERS; ulTemp++)
  {
    PktBuffers[ulTemp][1] = PktBuffers[ulTemp][0] = 0;
  }

  xEthRxInterruptQueue = xQueueCreate(emacNUM_RX_BUFFERS+1, sizeof(void *));
  xEthTxInterruptQueue = xQueueCreate(emacNUM_RX_BUFFERS+1, sizeof(void *));
}
