/*
    FreeRTOS.org V4.5.0 - Copyright (C) 2003-2007 Richard Barry.

    This file is part of the FreeRTOS.org distribution.

    FreeRTOS.org is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    A special exception to the GPL can be applied should you wish to distribute
    a combined work that includes FreeRTOS.org, without being obliged to provide
    the source code for any proprietary components.  See the licensing section
    of http://www.FreeRTOS.org for full details of how and when the exception
    can be applied.

    (c) 2012-2016 modified by J.Fojtik
*/
#include <stdlib.h>

/* Kernel includes. */
#include "FreeRTOS.h"
#include "Semphr.h"
#include "task.h"

/* Demo includes. */
#include "EMAC.h"

/* uIP includes. */
#include "uip.h"
#include "uip_arp.h"

/* Hardware library includes. */
#include "hw_types.h"
#include "hw_memmap.h"
#include "hw_ints.h"
#include "hw_ethernet.h"
#include "hw_timer.h"
#ifdef __TIMER_H__
 #undef __TIMER_H__
 #include "driverlib/timer.h"
#endif
#include "ethernet.h"
#include "interrupt.h"
#include "driverlib/udma.h"

#include "typedfs.h"


#define SYSCTL_PERIPH_TIMERfix	SYSCTL_PERIPH_TIMER2
#define TIMERfix_BASE		TIMER2_BASE
#define INT_TIMERfixA		INT_TIMER2A


//#define DEBUG_EMAC		/* Debug Emac only, not needed. */
//extern void LogEvent(UINT16 level, const char *text, ...);
//#define	TRAFFIC_EVENT		0x4000


#ifdef DEBUG_EMAC
#include <stdio.h>
#include <string.h>
/*#include "UART2.h"		/// !!!!!
#define LOG2STREAM  UART2	// !!!!!
static const char T_char = 'T';
static const char R_char = 'R';
static const char WAKEUP_char = '!';
static const char IRQ_char = '{';
static const char IRQe_char = '}';
//static const char Space_char = ' ';
static const char r_char = 'r';
static const char t_char = 't';
static const char e2_char = '2';
static const char e8_char = '8';
static const char e9_char = '9';
static const char m_char = 'm';
static const char p_char = 'p';
static const char q_char = '?';
static const char PROC_char = '[';
static const char PROCe_char = ']';*/

char Time[10];
#define DBG_PUTCHAR(c)  StreamPutchar(&UART2, c)

#else
 #define DBG_PUTCHAR(c)
#endif


/* ************** USER Tunable area *****************/
#define DMA_TRANSFER		///< DMA offloads CPU.
#define EARLY_SEND		///< Allow to speed up transfer, start transmit when FIFO is not fully filled with packet.

#define macNEGOTIATE_DELAY 	2000
#define macWAIT_ETHER		20
#define WAIT_FOR_DMA_MS		50
#define CHECK_PROCESS_MS	1000
/****************************************************/

//#define taskENTER_CRITICAL __disable_irq
//#define taskEXIT_CRITICAL  __enable_irq

#define CRITICAL_LOCK   __disable_irq
#define CRITICAL_UNLOCK __enable_irq


#ifndef DMA_TRANSFER
 #ifdef EARLY_SEND
  #warning EARLY_SEND will work only when DMA is enabled!
  #undef EARLY_SEND
 #endif
#endif


#define emacFRAM_SIZE_BYTES 		   2


unsigned short EMAC_ErrorFlags = 0;	/**< Bit array with error flags, see ::EMAC_ERRORS. */


#define EMAC_NEED_RESTART	1
#define EMAC_NEED_TX		2
#define EMAC_NEED_RX		4

UINT8 EmacNeedJobs = 0;			/**< 1 - Need restart, 2 - Need Tx processing, 4 - Need Rx packet pull */
static volatile UINT32 *pushBufferISR = NULL; /** Stores a pointer to packet, that is currently transmitted. */

volatile char TransmitInProgress = 0;	/**< Sets to 1 when a packet is really transmitted. */
#define MAX_TX_PROGRESS	5		/**< Maximal alloved ::TransmitInProgress. */

UINT16 JumboSize = 0;			/** Stores packet size of rejected jumbo. */
UINT32 TimemarkRxPacket;
UINT32 TimemarkTxPacket;
#define TX_TIMEOUTms		2	/** Timeout to wait for packet to be transmitted [ms]. */

#define MAC_PROCESSING_RX	1	/** Packet receiving is processed in external task - see ::MacJob */
#define MAC_PROCESSING_TX	2	/** Packet transmition is processed in external task - see ::MacJob */
#define MAC_TASKACTIVE		4	/** MAC processing job is running - see ::MacJob */
volatile unsigned char MacJob = 0;



void SetupDMA(void);
void prvSetMACAddress(void);


/* The task that handles the MAC peripheral.  This is created at a high
priority and is effectively a deferred interrupt handler.  The peripheral
handling is deferred to a task to prevent the entire FIFO having to be read
from within an ISR. */
void vMACHandleTask(void *pvParameters);

/*-----------------------------------------------------------*/

/* The semaphore used to wake the interrupt handler task.  The peripheral
is processed at the task level to prevent the need to read the entire FIFO from
within the ISR itself. */
xSemaphoreHandle xMACInterruptSemaphore = NULL;

/* The queue used to wake the uIP task when data arrives. */
extern xQueueHandle xEthRxInterruptQueue;
extern xQueueHandle xEthTxInterruptQueue;


/*-----------------------------------------------------------*/

extern void vuIP_Task(void *pvParameters);

/* Size of the stack allocated to the uIP task. */
#define mainBASIC_VUIP_STACK_SIZE	(configMINIMAL_STACK_SIZE + 200)

#define mainCHECK_TASK_PRIORITY		(tskIDLE_PRIORITY + 2)


/** Initialise low-level ethernet hardware. */
int EthInitModule(void)
{ 
  if(!SysCtlPeripheralPresent(SYSCTL_PERIPH_ETH)) return -1;
  InitPktPool();

	/* Enable/Reset the Ethernet Controller */
  SysCtlPeripheralEnable(SYSCTL_PERIPH_ETH);
  SysCtlPeripheralSleepEnable(SYSCTL_PERIPH_ETH);
  	/* Ensure all interrupts are disabled. */
  EthernetIntDisable(ETH_BASE, (ETH_INT_PHY|ETH_INT_MDIO|ETH_INT_RXER|ETH_INT_RXOF|ETH_INT_TX|ETH_INT_TXER|ETH_INT_RX));

  if(xTaskCreate(vuIP_Task, (signed portCHAR *)"uIP", mainBASIC_VUIP_STACK_SIZE, NULL, mainCHECK_TASK_PRIORITY-1, NULL) != pdPASS)
     return -2;

  	/* Create the queue and task used to defer the MAC processing to the task level. */
  if(xMACInterruptSemaphore==NULL)
  {
    vSemaphoreCreateBinary(xMACInterruptSemaphore);
    xSemaphoreTake(xMACInterruptSemaphore, 0);
  }
  if(xMACInterruptSemaphore==NULL) return -3;

  if(xTaskCreate(vMACHandleTask, (signed portCHAR *)"MAC", configMINIMAL_STACK_SIZE, NULL, configMAX_PRIORITIES-1, NULL ) != pdPASS)
	 return -4;     

#ifdef DMA_TRANSFER
  SetupDMA();						// HW setup of uDMA.

  	// Configure the DMA channel for Ethernet receive.
  uDMAChannelAttributeDisable(UDMA_CHANNEL_ETH0RX, UDMA_ATTR_ALL);
  uDMAChannelControlSet(UDMA_CHANNEL_ETH0RX, UDMA_SIZE_32 | UDMA_SRC_INC_NONE | UDMA_DST_INC_32 | UDMA_ARB_8);

	// Configure the DMA TX channel
  uDMAChannelAttributeDisable(UDMA_CHANNEL_ETH0TX, UDMA_ATTR_ALL);
  uDMAChannelControlSet(UDMA_CHANNEL_ETH0TX, UDMA_SIZE_32 | UDMA_SRC_INC_32 | UDMA_DST_INC_NONE | UDMA_ARB_8);

  uDMAEnable();
#endif

  HWREG(ETH_BASE+MAC_O_LED) = MAC_LED_LED1_LINKACT | MAC_LED_LED0_LINK;	// Enable EMAC LED diodes.

  	// Enable processor interrupts.    
  //IntMasterEnable();  

  return 0;
}


/*-----------------------------------------------------------*/

/** Init of busy-loop timer.
 * @param[in]	speedHz		Frequency of the safety timer in Hz.
 * @return      0 on success; otherwise nonzero on error  */
int SafetyTimer_Init(int speedHz)
{
  if(speedHz<=0) speedHz = 1;
  speedHz = SysCtlClockGet() / speedHz;

	/* Timer zero is used to measure time. */
  SysCtlPeripheralEnable(SYSCTL_PERIPH_TIMERfix);
 
  TimerConfigure(TIMERfix_BASE, TIMER_CFG_32_BIT_PER);

	/* Just used to measure time. */
  TimerLoadSet(TIMERfix_BASE, TIMER_A, speedHz);

	/* Enable timer 0. */	
  TimerEnable(TIMERfix_BASE, TIMER_A);  
  
  if(HWREG(TIMERfix_BASE + TIMER_O_TAV)==speedHz) return -1;	// timer is stucked.

	/* We are only interested in Rx interrupts. */
  IntPrioritySet(INT_TIMERfixA, configKERNEL_INTERRUPT_PRIORITY);
  //IntEnable(INT_TIMERfixA);

  TimerIntEnable(TIMERfix_BASE,TIMER_TIMA_TIMEOUT);
return 0;
}


/** Setup eth hardware peripheral. */
portBASE_TYPE vInitEMAC(void)
{
unsigned long ulTemp;

	/* Enable/Reset the Ethernet Controller */  
  SysCtlPeripheralReset(SYSCTL_PERIPH_ETH);

	/* Ensure all interrupts are disabled. */
  EthernetIntDisable(ETH_BASE, (ETH_INT_PHY|ETH_INT_MDIO|ETH_INT_RXER|ETH_INT_RXOF|ETH_INT_TX|ETH_INT_TXER|ETH_INT_RX));

	/* Clear any interrupts that were already pending. */
  ulTemp = EthernetIntStatus(ETH_BASE, pdFALSE);
  EthernetIntClear(ETH_BASE, ulTemp);

	/* Initialise the MAC and connect. */
  EthernetInit(ETH_BASE);
  EthernetConfigSet(ETH_BASE, (ETH_CFG_TX_DPLXEN|ETH_CFG_TX_CRCEN|ETH_CFG_TX_PADEN));
  EthernetEnable(ETH_BASE);  
  
  vTaskDelay(ms2tickCount(macNEGOTIATE_DELAY));
	
	/* We are only interested in Rx interrupts. */
  IntPrioritySet(INT_ETH, configKERNEL_INTERRUPT_PRIORITY);
  IntEnable(INT_ETH);
  MacJob = 0;  
  EthernetIntEnable(ETH_BASE, ETH_INT_RX|ETH_INT_TX);

  SafetyTimer_Init(1000);

return pdPASS;
}


/*-----------------------------------------------------------*/


/** Fill data received to ucBuffer.
 * @param[out] ucBuffer		Buffer to get a packet received.
 * @param[in] Timeout		Timeout in systicks. */
unsigned int uiGetEMACRxData(unsigned char **ucBuffer, int Timeout)
{
unsigned char *uip_buf2;
unsigned uip_len2;

 if(ucBuffer==NULL) return 0;
 if(xQueueReceive(xEthRxInterruptQueue, &uip_buf2, Timeout))
 {
   if(uip_buf2!=NULL)			// Some packet has been received.
   {
     uip_len2 = *(WORD *)uip_buf2;
     if(uip_len2<=4)			// considered to be invalid packet
     {
	DisposePkt(uip_buf2,0);		// release an invalid packet
	uip_buf2 = NULL;
     }
     else
     {
	uip_len2 -= 2;
	DisposePkt(*ucBuffer,0);	// Release old buffer.
	*ucBuffer = uip_buf2 + 2;
        return uip_len2;
     }
   }
 }

return 0;		// no new valid packet received
}


/*-----------------------------------------------------------*/



/** Fix stucked transmit from external timer. */
void fTimer2A_ISR(void)
{  
  TimerIntClear(TIMERfix_BASE,0xFFFFFFFF);

  if(TransmitInProgress == 0)	/* No active transmition - stop the Tx watch timer. */
  {
    IntDisable(INT_TIMERfixA);
    return;
  }

  if(TransmitInProgress > MAX_TX_PROGRESS)
  {
    IntDisable(INT_TIMERfixA);
    TransmitInProgress = 0;    
    return;
  }

	// Check here whether Tx IRQ has been lost.
	// When Tx IRQ is fired, pushBufferISR is cleaned out inside this IRQ silently.
    	// When everything works as expected, this code should not occur.
#ifdef DMA_TRANSFER
    if(pushBufferISR)					// Tx buffer contains data?
    {
      if(!uDMAChannelIsEnabled(UDMA_CHANNEL_ETH0TX))	// DMA transfer finished?
      {
        if(EthernetSpaceAvail(ETH_BASE))
            HWREG(ETH_BASE+MAC_O_TR) = MAC_TR_NEWTX;	/* Fire Tx packet. */
#ifdef DEBUG_EMAC
        xQueueSendFromISR(LOG2STREAM.Tx_QueueHandle, &T_char, &xSwitchRequired);
#endif      
        DisposePkt((UINT8*)pushBufferISR,1);		// Packet is deallocated when copyed to TX FIFO.
        pushBufferISR = NULL;      
        return;
      }    
    }
#endif


  if(labs(xTaskGetTickCountFromISR()-TimemarkTxPacket) > ms2tickCount(TX_TIMEOUTms))
  {
    TimemarkTxPacket += 1;

    if(!uDMAChannelIsEnabled(UDMA_CHANNEL_ETH0TX))
      if(EthernetSpaceAvail(ETH_BASE))
      {
        TransmitInProgress++;
        EMAC_ErrorFlags |= EMAC_ERR_TX_RETRANSMIT;
        HWREG(ETH_BASE+MAC_O_TR) = MAC_TR_NEWTX;	/* Retransmit a packet. */
      }
  }
}


/** This function sends a packet pointed by ppus. If Tx opperation is in progress it allocates new buffer into ppus.
 * @param[in,out]   ppus	pointer to packet PTR. PACKET + 2byte MUST BE 32bit alligned and THERE MUST BE 2 BYTES RESERVED
 *                        BEFORE PACKET. Thus packet - 2B must be 32 bit alligned. This is because DMA transfer needs this.
 *                        When packet is moved to queue, *ppus is set to NULL. Otherwise it is left intact.
 * @param[in]	ulNextTxSpace	Full size of packet including ETH and IPv4 header. */
void vSendBufferToMAC(unsigned char **ppus, portSHORT ulNextTxSpace)
{
unsigned char *pus;
int Timeout;

//LogEvent(TRAFFIC_EVENT, "Send packet - len=%d, mss=%d conn:%d", ulNextTxSpace, uip_conn->mss, uip_conn-uip_conns);
//printf("\r\nSend packet - len=%d, mss=%d conn:%d", ulNextTxSpace, uip_conn->mss, uip_conn-uip_conns);
  //printf("T");

#ifdef DEBUG_EMAC
  if(TransmitInProgress) DBG_PUTCHAR('~');
                     else DBG_PUTCHAR('-');
#endif 

  if(ppus==NULL) return;
  if(*ppus==NULL) return;
  if(ulNextTxSpace<=14) return;

	/* Locate the data to be send. */
  pus = *ppus;
	/* Add in the size of the data. */  
  pus -= 2;					/* Block size is stored 2 bytes *BEFORE* pointer pus! */
  *(UINT16*)pus = ulNextTxSpace - 14; // + 2;	/* First 14 bytes are not included inside packet length */

#ifdef DMA_TRANSFER
  if((MacJob&MAC_PROCESSING_TX)!=0 			// External task is processing TX?
      || uDMAChannelIsEnabled(UDMA_CHANNEL_ETH0TX) 	// DMA Tx is running?
      || !EthernetSpaceAvail(ETH_BASE)			// ETH is transmitting data?
      || TransmitInProgress
//      || pushBufferISR!=NULL 				// Waiting for finishing Tx cycle?      
    )
  {
    if(xQueueSendToBack(xEthTxInterruptQueue,&pus,1) == pdPASS)
    {
      *ppus = NULL;		// Toss away uIP packet.
      return;			// Packet will be handled inside ISR.
    }

		// Adding to a queue failed, try to continue normally.  
    Timeout = 0;
    while(uDMAChannelIsEnabled(UDMA_CHANNEL_ETH0TX))	// Wait for ETH Tx DMA to finish
    {
      if(Timeout++>WAIT_FOR_DMA_MS)
	{EMAC_ErrorFlags|=EMAC_ERR_DMA; break;}	// DMA timeout
      vTaskDelay(1);
    }  
  }

  if(xIsQueueEmpty(xEthTxInterruptQueue)==pdFALSE)	// Is something in Tx FIFO?
  {
    if(xQueueSendToBack(xEthTxInterruptQueue,&pus,1) == pdPASS)
    {
      *ppus = NULL;		// Toss away uIP packet.
      xSemaphoreGive(xMACInterruptSemaphore);	// Looks bad, try to fire EMAC task.
      return;			// Packet will be handled inside ISR.
    }
  }
 
#endif /* DMA_TRANSFER */

	/* Wait for data to be sent if there is no space immediately. */
  Timeout = 0;
  while(!EthernetSpaceAvail(ETH_BASE))		// Tady by mohlo zacit kopirovani drive OVERIT volne misto v Tx ETH!!!!!
  {
    if(Timeout++>macWAIT_ETHER)
    {
      if(EMAC_ErrorFlags & EMAC_ERR_ETHER) EmacNeedJobs |= EMAC_NEED_RESTART;
      EMAC_ErrorFlags |= EMAC_ERR_ETHER;
      break;			// Ethernet timeout
    }
    vTaskDelay(1);
  }  	

#ifdef DMA_TRANSFER  

  taskENTER_CRITICAL();				/* ***** CRITICAL SECTION ******* */ 
  if((MacJob&MAC_PROCESSING_TX)!=0)
  {
    taskEXIT_CRITICAL();			/* We are in big trouble, packet cannot be sent - toss it away. */
    if(xQueueSendToBack(xEthTxInterruptQueue,&pus,0) == pdPASS)    
      *ppus = NULL;				/* Enforce caller to allocate new packet */
    else    
      EMAC_ErrorFlags |= EMAC_ERR_TX_OVERFLOW;	/* *ppus can be re-used, signalise error. */
    return;
  }
  MacJob |= MAC_PROCESSING_TX; 
  taskEXIT_CRITICAL();				/* ~~~~~~ CRITICAL SECTION ~~~~~~ */

  //EthernetIntDisable(ETH_BASE, ETH_INT_RX|ETH_INT_TX);  // This is not needed, because of MAC_PROCESSING_TX flag
  if(pushBufferISR!=NULL)			/* DMA channel finished and pushBufferISR is stucked? */
  {
    DisposePkt((UINT8*)pushBufferISR,0);	/* Release pushBufferISR. */
    pushBufferISR = NULL;
    EMAC_ErrorFlags |= EMAC_ERR_ST_PKT;
  }

  uDMAChannelTransferSet(UDMA_CHANNEL_ETH0TX, UDMA_MODE_AUTO, pus, (void *)(ETH_BASE+MAC_O_DATA), (ulNextTxSpace+2 +3)>>2);
  uDMAChannelEnable(UDMA_CHANNEL_ETH0TX);

  pushBufferISR = (UINT32*)*ppus;
  *ppus = NULL;					// Toss away uIP packet
  
#ifdef EARLY_SEND
  if(ulNextTxSpace<40 || SysCtlClockGet()<70000000)
    HWREG(ETH_BASE + MAC_O_THR) = 0x3F;	// Disable autotransmit.  ----- !!!!! TODO: overit popripade zrusit.
  else  
    HWREG(ETH_BASE + MAC_O_THR) = 1;	// Enable autotransmit when packet is larger than 40byte. 4*(THR*8 + 1)      
#endif	// ~EARLY_SEND

     ////////////////////////////////////////////////////
  taskENTER_CRITICAL();			// **** Critical section, no job can interfere ****  		  
  MacJob &= ~MAC_PROCESSING_TX;  		// Clear Tx ussage flag.

  if(!EthernetSpaceAvail(ETH_BASE))	// Big trouble again some packet starts processing asynchronously before lock.
  {
    if(xQueueSendToBack(xEthTxInterruptQueue,&pushBufferISR,0) == pdPASS)
      pushBufferISR = NULL;
    taskEXIT_CRITICAL();
    return;
  }

  TransmitInProgress = 1;
  if(MacJob==0)
  {
    //EthernetIntEnable(ETH_BASE, ETH_INT_RX|ETH_INT_TX);  // This is not needed, because of MAC_PROCESSING_TX flag    
    uDMAChannelRequest(UDMA_CHANNEL_ETH0TX);	// Start DMA transfer, after IRQ is enabled.    
    TimemarkTxPacket = xTaskGetTickCount();    
  }
  else
  {
    uDMAChannelRequest(UDMA_CHANNEL_ETH0TX);	// Start DMA transfer.
    TimemarkTxPacket = xTaskGetTickCount();
    xSemaphoreGive(xMACInterruptSemaphore);	// Wakeup MAC proces.
  }
  IntEnable(INT_TIMERfixA);
  taskEXIT_CRITICAL();				// **** ~~Critical section, no job can interfere~~ ****
     ////////////////////////////////////////////////////

  DBG_PUTCHAR('@');
  return;		// Return here, ISR will start transfer when DMA finish transfer to EMAC FIFO.

#else /* ~DMA_TRANSFER */

  EthernetIntDisable(ETH_BASE, ETH_INT_RX|ETH_INT_TX);
  if(pushBufferISR!=NULL)			/* DMA channel finished and pushBufferISR is stucked? */
  {
    DisposePkt((UINT8*)pushBufferISR,0);	/* Release pushBufferISR. */
    pushBufferISR = NULL;
    EMAC_ErrorFlags |= EMAC_ERR_ST_PKT;
  }

  DBG_PUTCHAR(T_char);

	/* Copy packet to EMAC FIFO without DMA support. */
  {
    UINT32 *pulSource;
    pulSource = (unsigned portLONG *)pus;
    ulNextTxSpace =  (ulNextTxSpace+2 +3)>>2;		// Amount of DWORDs to be transmitter.
    while(ulNextTxSpace>0)
    {
      HWREG(ETH_BASE + MAC_O_DATA) = *pulSource;
      pulSource++;
      ulNextTxSpace--;
    }  
  }	
  HWREG(ETH_BASE + MAC_O_TR) = MAC_TR_NEWTX;		/* Go Tx packet. */ 
  /* ppus is still valid and can be re-used, packet is fully copied into TX FIFO. */
#endif
}


/*-----------------------------------------------------------*/

/** IRQ from ethernet device.
 * Global variable - pushBufferISR  packet aleready beeing sent. This value is cleared when send succeeds. */
void vEMAC_ISR(void)
{
portBASE_TYPE xSwitchRequired = pdFALSE;
unsigned portLONG ulTemp;
static UINT32 *pulBufferISR = NULL;
unsigned ulLength;

  CpuLoadEnable();

	/* Clear the interrupt. */
  ulTemp = EthernetIntStatus(ETH_BASE, pdFALSE);
  EthernetIntClear(ETH_BASE, ulTemp);

  if(getCriticalNesting())		// RTOS failure, IRQ should be disabled, bail out.
  {
    EMAC_ErrorFlags |= EMAC_ERR_BAD_IRQ;
    goto ExitISR;
  }

#ifdef DEBUG_EMAC
    ///////  !!!!!!!!!!!!!!!!	VYHODIT !!!!!!!!!!!!!!!!!!      
  //xQueueSendFromISR(LOG2STREAM.Tx_QueueHandle, &IRQ_char, &xSwitchRequired);

/*    sprintf(Time,"%d",xTaskGetTickCountFromISR());
    for(ulLength=0;ulLength<sizeof(Time);ulLength++)
    {
      if(Time[ulLength]==0) break;
      xQueueSendFromISR(LOG2STREAM.Tx_QueueHandle, &Time[ulLength], &xSwitchRequired);
    } */

  //if(ulTemp & ETH_INT_RX)
  //    {xQueueSendFromISR(LOG2STREAM.Tx_QueueHandle, &r_char, &xSwitchRequired);}
  if(ulTemp & ETH_INT_TXER)
      {xQueueSendFromISR(LOG2STREAM.Tx_QueueHandle, &e2_char, &xSwitchRequired);}
  if(ulTemp & ETH_INT_TX) 
      {xQueueSendFromISR(LOG2STREAM.Tx_QueueHandle, &t_char, &xSwitchRequired);}
  if(ulTemp & ETH_INT_RXOF)
      {xQueueSendFromISR(LOG2STREAM.Tx_QueueHandle, &e8_char, &xSwitchRequired);}
  if(ulTemp & ETH_INT_RXER)
      {xQueueSendFromISR(LOG2STREAM.Tx_QueueHandle, &e9_char, &xSwitchRequired);}
  if(ulTemp & ETH_INT_MDIO)
      {xQueueSendFromISR(LOG2STREAM.Tx_QueueHandle, &m_char, &xSwitchRequired);}
  if(ulTemp & ETH_INT_PHY)
      {xQueueSendFromISR(LOG2STREAM.Tx_QueueHandle, &p_char, &xSwitchRequired);}

  //if(ulTemp==0)
  //    xQueueSendFromISR(LOG2STREAM.Tx_QueueHandle, &q_char, &xSwitchRequired);
  ///////  !!!!!!!!!!!!!!!!	VYHODIT !!!!!!!!!!!!!!!!!! */
#endif // DEBUG_EMAC

  if(ulTemp & ETH_INT_TXER)
  {    
    EMAC_ErrorFlags |= EMAC_ERR_TX_ERROR;    
  }

  if(ulTemp & ETH_INT_RXOF)		// Rx FIFO overrun.
  {   
    HWREG(ETH_BASE+MAC_O_RCTL) &= ~MAC_RCTL_RXEN;   // Rx disable
    HWREG(ETH_BASE+MAC_O_RCTL) |= MAC_RCTL_RSTFIFO; // Clear FIFO

    ulTemp &= ~ETH_INT_RX;		// Clear receive packet flag
    EMAC_ErrorFlags |= EMAC_ERR_RX_OVERRUN;
    HWREG(ETH_BASE+MAC_O_RCTL) |= MAC_RCTL_RXEN;	// Rx enable
  }

  if((ulTemp & ETH_INT_TX) != 0)	// Transmition has been finished.
  {
    IntDisable(INT_TIMERfixA);		// No need to check for transmition stuck.
    TransmitInProgress = 0;
  }

  if((ulTemp & ETH_INT_RX)==0)
  {
    if((HWREG(ETH_BASE+MAC_O_NP)&0x3F)>=1)		// If there is any packet in a queue,
    {
      ulTemp |= ETH_INT_RX;
    }
  } 

#ifdef DMA_TRANSFER
  if(pushBufferISR)					// Tx buffer contains data?
  {
    if(!uDMAChannelIsEnabled(UDMA_CHANNEL_ETH0TX))	// DMA transfer finished?
    {
#ifdef EARLY_SEND
      if(HWREG(ETH_BASE+MAC_O_THR)==0x3F)		// Fire packet only when autotransmit is NOT set.
#endif	// EARLY_SEND
      {    	  
        HWREG(ETH_BASE+MAC_O_TR) = MAC_TR_NEWTX;	/* Go Tx packet. */
#ifdef DEBUG_EMAC
        xQueueSendFromISR(LOG2STREAM.Tx_QueueHandle, &T_char, &xSwitchRequired);
#endif
      }

      DisposePkt((UINT8*)pushBufferISR,1);		// Packet is deallocated when copyed to TX FIFO.
      pushBufferISR = NULL;      
    }    
  }

  if(pushBufferISR==NULL)				// No packet is copying to TX FIFO?
  {
    if(EthernetSpaceAvail(ETH_BASE))			// Is Tx FIFO free?
    {
      if((MacJob&MAC_PROCESSING_TX)==0)			// Is any external task that process Transmitting.
      {
        if(TransmitInProgress==0)
        {
          if(xQueueReceiveFromISR(xEthTxInterruptQueue, &pushBufferISR, &xSwitchRequired) == pdTRUE)	// look for other packet to send
          {
            if(pushBufferISR!=NULL)
            {						// Get a new packet and feed it into TxFIFO using DMA.          
              uDMAChannelTransferSet(UDMA_CHANNEL_ETH0TX, UDMA_MODE_AUTO, (void*)pushBufferISR, (void*)(ETH_BASE+MAC_O_DATA), (*(UINT16*)pushBufferISR+2 +3)>>2);
              uDMAChannelEnable(UDMA_CHANNEL_ETH0TX);
#ifdef EARLY_SEND
              if(*(UINT16*)pushBufferISR<40)
                HWREG(ETH_BASE + MAC_O_THR) = 0x3F;	// Disable autotransmit.
              else
                HWREG(ETH_BASE + MAC_O_THR) = 1;	// Enable autotransmit when packet is larger than 40byte.
#endif	/* ~EARLY_SEND */
              TransmitInProgress = 1;
              uDMAChannelRequest(UDMA_CHANNEL_ETH0TX);
              TimemarkTxPacket = xTaskGetTickCountFromISR();
              IntEnable(INT_TIMERfixA);
            }        
          }
          else
            pushBufferISR = NULL;

          ulTemp &= ~ETH_INT_TX;				// Transmit is dispatched.
        }
      }      
    }
  }


  if(!uDMAChannelIsEnabled(UDMA_CHANNEL_ETH0RX))  // Continue this only if DMA is not running.
  {
    if(pulBufferISR!=NULL)			//DMA transfer finished
    {
      if((MacJob&MAC_PROCESSING_RX)==0)
        EthernetIntEnable(ETH_BASE, ETH_INT_RX);	// Enable Rx interrupt, when Rx DMA finished, it emits IRQ with signature 0.
      else
        xSemaphoreGiveFromISR(xMACInterruptSemaphore, &xSwitchRequired);
        
		/* Ensure the uIP task is not blocked as data has arrived. */
      if(errQUEUE_FULL==xQueueSendFromISR(xEthRxInterruptQueue, &pulBufferISR, &xSwitchRequired))
      {
	DisposePkt((UINT8*)pulBufferISR,1);
        EMAC_ErrorFlags |= EMAC_ERR_RX_OVERFLOW;
      }
      pulBufferISR = NULL;
      
      TimemarkRxPacket = xTaskGetTickCountFromISR();

      //if((HWREG(ETH_BASE+MAC_O_NP)&0x3F)>=1)		// If there is any packet in a queue,
        //ulTemp |= ETH_INT_RX;				// turn Rx flag on to process it immediatelly.
    }    

    if(ulTemp & ETH_INT_RX)				/* Was it an Rx interrupt? */
    {
      if((HWREG(ETH_BASE + MAC_O_NP)&0x3F)>=1)		// Is there any packet received?
      {
        ulLength = HWREG(ETH_BASE + MAC_O_DATA);	// Read frame size

        if((ulLength&0xFFFF)>4)
        {
          if((ulLength&0xFFFF)<=UIP_BUFSIZE+8)		// UIP buffer has padding 8 bytes
          {	
            pulBufferISR = (UINT32 *)GetFreePkt(1);

            if(pulBufferISR==NULL)
            {	// Toss away FIFO content with DMA dummy read.	    	    
              HWREG(ETH_BASE+MAC_O_RCTL) &= ~MAC_RCTL_RXEN;	// Rx disable
	      HWREG(ETH_BASE+MAC_O_RCTL) |= MAC_RCTL_RSTFIFO;	// Reset receive FIFO.
              EMAC_ErrorFlags |= EMAC_ERR_OVERLOAD;
              HWREG(ETH_BASE+MAC_O_RCTL) |= MAC_RCTL_RXEN;	// Rx enable              
            }
            else
            {		/* Leave room at the start of the buffer for the size. */
	      *pulBufferISR = ulLength;
	      ulLength &= 0xFFFF;
	      ulLength -= 4;	  
		// Set up the DMA to transfer the Ethernet header when a packet is received    
              uDMAChannelTransferSet(UDMA_CHANNEL_ETH0RX, UDMA_MODE_AUTO, (void *)(ETH_BASE+MAC_O_DATA), pulBufferISR+1, (ulLength+3)>>2);
	      uDMAChannelEnable(UDMA_CHANNEL_ETH0RX);
      		// Issue a software request to start the channel running.    
              uDMAChannelRequest(UDMA_CHANNEL_ETH0RX);

              EthernetIntDisable(ETH_BASE, ETH_INT_RX);	//Disable IRQ from Rx here, DMA will emit IRQ with no signature.
            }                
	  }      
          else		// Jumbo packet received :(
          {
            JumboSize = ulLength & 0xFFFF;
	    HWREG(ETH_BASE+MAC_O_RCTL) &= ~MAC_RCTL_RXEN;	// Rx disable
	    HWREG(ETH_BASE+MAC_O_RCTL) |= MAC_RCTL_RSTFIFO;	// Reset receive FIFO.
            EMAC_ErrorFlags |= EMAC_ERR_JUMBO;
            HWREG(ETH_BASE+MAC_O_RCTL) |= MAC_RCTL_RXEN;	// Rx enable	  
	  }
        }
        else 
        {				// No Reset needed
	  EMAC_ErrorFlags |= EMAC_ERR_TOOSMALL;	
        }
      
        ulTemp &= ~ETH_INT_RX;        
						   	//The Rx IRQ will be re-enabled after DMA finish.
      }
      else	// (HWREG(ETH_BASE + MAC_O_NP)&0x3F)>=1
        ulTemp &= ~ETH_INT_RX;				// No packet received and Rx flag is up, too bad!
    }    
  }

#else /* ~DMA_TRANSFER */

  if(pushBufferISR==NULL)			// TX finished
  {
    if(EthernetSpaceAvail(ETH_BASE))
      if(xQueueIsQueueEmptyFromISR(xEthTxInterruptQueue)==pdFALSE)	// Is there any other Tx packet waiting
      {
        xSemaphoreGiveFromISR(xMACInterruptSemaphore, &xSwitchRequired);
      }
  }

#endif /* DMA_TRANSFER */


	/* Was there an Rx interrupt unserviced?
	   Just for a DMA case, if DMA is still running, switch to a specialised task. */
  if((ulTemp & ETH_INT_RX) != 0)
  {    
    xSemaphoreGiveFromISR(xMACInterruptSemaphore, &xSwitchRequired);
    EthernetIntDisable(ETH_BASE, ETH_INT_RX);
    EmacNeedJobs |= EMAC_NEED_RX;		// Set MACJOB
#ifdef DMA_TRANSFER				// This is not an error when DMA is disabled.
    EMAC_ErrorFlags |= EMAC_ERR_RX_ISR_NACC;
#endif
  }
  if((ulTemp & ETH_INT_TX) != 0)
  {
    if(xQueueIsQueueEmptyFromISR(xEthTxInterruptQueue)==pdFALSE)	// Is there any other Tx packet waiting?
    {
      EmacNeedJobs |= EMAC_NEED_TX;
      xSemaphoreGiveFromISR(xMACInterruptSemaphore, &xSwitchRequired);	// Wakeup MAC process to receive TX new packet
    }
  }

ExitISR:

#ifdef DEBUG_EMAC
  //xQueueSendFromISR(LOG2STREAM.Tx_QueueHandle, &IRQe_char, &xSwitchRequired);
#endif

	/* Switch to the uIP task. */
  portEND_SWITCHING_ISR(xSwitchRequired);  
}


/*-----------------------------------------------------------*/

/** This FreeRTOS process acts as a watchdog to prevent ETH DMA stuck. */
void vMACHandleTask(void *pvParameters)
{
UINT32 ulLength;
UINT32 *pulBufferCP;
UINT8  *pulBuffer;

#ifdef DEBUG_EMAC
  UART2_DriverInit(115200,0,0);
#endif

  pulBuffer = NULL;
  for( ;; )
  {
#ifdef DEBUG_EMAC
    UART2_ActivateTransfer();
#endif
		/* Wait for something to do. */
    if(xSemaphoreTake(xMACInterruptSemaphore, CHECK_PROCESS_MS)!=pdTRUE)
    {		/* When timeout occurs, check DMA status first. */
      if(EmacNeedJobs==0)
      {
#ifdef DMA_TRANSFER
        if(uDMAChannelIsEnabled(UDMA_CHANNEL_ETH0RX)) continue;		// If Rx DMA running, ignore wakeup.
#endif
        if(labs(xTaskGetTickCount()-TimemarkRxPacket) < ms2tickCount(1000))
	    continue;		// When some packet arrived, the interface is not hanged.
      }
    }
    else
    {
      DBG_PUTCHAR(WAKEUP_char);
    }

    CRITICAL_LOCK(); MacJob |= MAC_TASKACTIVE; CRITICAL_UNLOCK();		// Nobody should allow IRQ
    EthernetIntDisable(ETH_BASE, ETH_INT_RX|ETH_INT_TX);

    DBG_PUTCHAR(PROC_char);
	
    if(EmacNeedJobs & EMAC_NEED_RESTART)  /* Ethernet device seems to crash. Try to reinitialise it. */
    {			
      EmacNeedJobs = 0;
      vInitEMAC();
      prvSetMACAddress();
      //EthernetIntEnable(ETH_BASE, ETH_INT_RX); - INT enable is already called inside vInitEMAC.
      EMAC_ErrorFlags |= EMAC_ERR_RESTARTED;      
      continue;
    }

#ifdef DMA_TRANSFER
    if(!uDMAChannelIsEnabled(UDMA_CHANNEL_ETH0RX))
#endif
    {
      CRITICAL_LOCK(); MacJob |= MAC_PROCESSING_RX; CRITICAL_UNLOCK();

      //while(((EthernetIntStatus(ETH_BASE,pdFALSE) & ETH_INT_RX )) != 0)
      while((HWREG(ETH_BASE+MAC_O_NP)&0x3F)>=1)
      {		
        ulLength = HWREG(ETH_BASE + MAC_O_DATA);        

	  /* Assign a buffer only if data will fit in our buffer.  Ensure we don't try to write into the buffer. */
	if((ulLength&0xFFFF)<=4)
	{
	  EMAC_ErrorFlags |= EMAC_ERR_TOOSMALL;
	  continue;
	}
	
        if((ulLength&0xFFFF)<=UIP_BUFSIZE+8)	// UIP buffer has padding 8 bytes.
        {
          pulBuffer = GetFreePkt(0);      		          

          if(pulBuffer==NULL)
          {
	    HWREG(ETH_BASE+MAC_O_RCTL) |= MAC_RCTL_RSTFIFO;	// Reset receive FIFO.
	    EMAC_ErrorFlags |= EMAC_ERR_OVERLOAD;
	    break;
          }
          else
          {
            pulBufferCP = (UINT32*)pulBuffer;		// Read packet size from EMAC
            *pulBufferCP++ = ulLength;	    
	
	    ulLength &= 0xFFFF;	/* Get the size of the data. */ 
            ulLength -=4;

#ifdef DMA_TRANSFER
		// Set up the DMA to transfer the Ethernet header when a packet is received    
            uDMAChannelTransferSet(UDMA_CHANNEL_ETH0RX, UDMA_MODE_AUTO, (void *)(ETH_BASE+MAC_O_DATA), pulBufferCP, (ulLength+3)>>2);
	    uDMAChannelEnable(UDMA_CHANNEL_ETH0RX);
      		// Issue a software request to start the channel running.    
            uDMAChannelRequest(UDMA_CHANNEL_ETH0RX);

	    ulLength = 0;
	    while(uDMAChannelIsEnabled(UDMA_CHANNEL_ETH0RX))
	    {
	      if(ulLength++>9000) break;	// DMA transfer timeout
	    }
#else
		/* Read out the data into our buffer. */	    
	    while(ulLength>=4)
	    {   
	      *pulBufferCP = HWREG(ETH_BASE + MAC_O_DATA);
	      pulBufferCP++;
	      ulLength -= sizeof(*pulBufferCP);		// UINT32
	    }
	    if(ulLength>0)			// 1,2 or 3 bytes remaining, copy the tail.
	    {
	      UINT32 Data;
              Data = HWREG(ETH_BASE + MAC_O_DATA);	      
              memcpy(pulBufferCP, &Data, ulLength);	                    
	    }

#endif

            DBG_PUTCHAR(R_char);

		/* Ensure the uIP task is not blocked as data has arrived. */
	    if(xQueueSendToBack(xEthRxInterruptQueue,&pulBuffer,1) != pdPASS)
	      DisposePkt(pulBuffer,0);
	    pulBuffer = NULL;
	  }
	}
        else
	{						  // Jumbo packet received :(
	  JumboSize = ulLength & 0xFFFF;
	  HWREG(ETH_BASE+MAC_O_RCTL) &= ~MAC_RCTL_RXEN;	  // Rx disable
          HWREG(ETH_BASE+MAC_O_RCTL) |= MAC_RCTL_RSTFIFO; // Reset receive FIFO.
	  EMAC_ErrorFlags |= EMAC_ERR_JUMBO;	  
          HWREG(ETH_BASE+MAC_O_RCTL) |= MAC_RCTL_RXEN;	  // Rx enable	  
	}      	
      }

      CRITICAL_LOCK(); EmacNeedJobs&=~EMAC_NEED_RX; MacJob&=~MAC_PROCESSING_RX; CRITICAL_UNLOCK();
     }


	/* TX queue check for stucking. */
#ifdef DMA_TRANSFER
    if(!uDMAChannelIsEnabled(UDMA_CHANNEL_ETH0TX))
#endif /* ~DMA_TRANSFER */
      if(EthernetSpaceAvail(ETH_BASE) && TransmitInProgress==0)
      {
        taskENTER_CRITICAL(); 
        if((MacJob & MAC_PROCESSING_TX)!=0) goto TxProcessingLocked;		// Prevent using Tx from 2 separate threads
        MacJob |= MAC_PROCESSING_TX; 
        taskEXIT_CRITICAL();

        if(pushBufferISR)					/* Is some packet waiting to transmit? */
        {         
#ifdef EARLY_SEND
          if(HWREG(ETH_BASE+MAC_O_THR)==0x3F)			// Go only when autotransmit is NOT set.
#endif	// EARLY_SEND
          {		
            HWREG(ETH_BASE+MAC_O_TR) = MAC_TR_NEWTX;		/* Go Tx packet to wire. */
            DBG_PUTCHAR(T_char);
      	  }
          DisposePkt((UINT8*)pushBufferISR,1);
          pushBufferISR = NULL;
        }
	else
          while(xQueueReceive(xEthTxInterruptQueue,&pushBufferISR,0) == pdTRUE)	// look for other packet to send
          {
            if(pushBufferISR!=NULL)
            {
	      int Timeout;
#ifdef DMA_TRANSFER
              uDMAChannelTransferSet(UDMA_CHANNEL_ETH0TX, UDMA_MODE_AUTO, (void*)pushBufferISR, (void *)(ETH_BASE+MAC_O_DATA), (*(UINT16*)pushBufferISR +2 +3)>>2);
              uDMAChannelEnable(UDMA_CHANNEL_ETH0TX);
#ifdef EARLY_SEND
              if(*(UINT16*)pushBufferISR<40)
                HWREG(ETH_BASE+MAC_O_THR) = 0x3F;	// Disable autotransmit.
              else
                HWREG(ETH_BASE+MAC_O_THR) = 1;		// Enable autotransmit when packet is larger than 40byte.
#endif	/* ~EARLY_SEND */
              TransmitInProgress = 1;
              uDMAChannelRequest(UDMA_CHANNEL_ETH0TX);              
              TimemarkTxPacket = xTaskGetTickCount();
              IntEnable(INT_TIMERfixA);

			/* Wait for DMA to finish transfer. */
              Timeout = 0;
              while(uDMAChannelIsEnabled(UDMA_CHANNEL_ETH0TX))	// Wait for ETH Tx DMA to finish
              {
                if(Timeout++>WAIT_FOR_DMA_MS)
	           {EMAC_ErrorFlags|=EMAC_ERR_DMA; break;}	// DMA timeout
                vTaskDelay(1);
              }              

#ifdef EARLY_SEND
              if(HWREG(ETH_BASE+MAC_O_THR)==0x3F)		// Go only when autotransmit is not set.
#endif	/* ~EARLY_SEND */
              {		
                 HWREG(ETH_BASE+MAC_O_TR) = MAC_TR_NEWTX;	/* Go Tx packet. */
                 DBG_PUTCHAR(T_char);
      	      }

              DisposePkt((UINT8*)pushBufferISR,0);		// Packet is fully copied to Tx FIFO
              pushBufferISR = NULL;             

#else	/* DMA_TRANSFER */
			/* Copy packet to EMAC FIFO without DMA support. */
              UINT32 *pushSource;
              ulLength = (*(UINT16*)pushBufferISR +2 +3)>>2;

  	      pushSource = (UINT32 *)(pushBufferISR + 2);
              while(ulLength>0)
              {
                HWREG(ETH_BASE + MAC_O_DATA) = *pushSource;
                pushSource++;
                ulLength --;
              }              

              HWREG(ETH_BASE + MAC_O_TR) = MAC_TR_NEWTX;		/* Go Tx packet. */

              DisposePkt((UINT8*)pushBufferISR,0);
              pushBufferISR = NULL;

#endif	/* ~DMA_TRANSFER */

			/* Wait for data to be sent if there is no space immediately. */
              Timeout = 0;
              while(!EthernetSpaceAvail(ETH_BASE))		// Tady by mohlo zacit kopirovani drive OVERIT volne misto v ETH!!!!!
              {
                if(Timeout++>macWAIT_ETHER)
                {
                  if(EMAC_ErrorFlags & EMAC_ERR_ETHER) EmacNeedJobs|=EMAC_NEED_RESTART;
                  EMAC_ErrorFlags |= EMAC_ERR_ETHER;
                  break;			// Ethernet timeout
                }
                vTaskDelay(1);
              }

              TransmitInProgress = 0;		// The Tx IRQ is just disabled, we must clear this flag manually.
              IntDisable(INT_TIMERfixA);
	    }
          }

        taskENTER_CRITICAL();
        MacJob &= ~MAC_PROCESSING_TX;
        EmacNeedJobs&=~EMAC_NEED_TX;
TxProcessingLocked:
        taskEXIT_CRITICAL();        
      }    
    
    CRITICAL_LOCK();
    MacJob &= ~MAC_TASKACTIVE;
    if(MacJob==0)
      EthernetIntEnable(ETH_BASE, ETH_INT_RX|ETH_INT_TX);
    CRITICAL_UNLOCK();

    DBG_PUTCHAR(PROCe_char);
  }
}


/* --------------------------------------------------------------------------*/

/* MAC is set from config.xml: <LAN> <MAC val="12:34:56:78:9A"> </LAN> */


void prvSetMACAddress(void)
{
  EthernetMACAddrSet(ETH_BASE, uip_ethaddr.addr);	/* Program the MAC address. */
}


